Unit AdLibFM;

(*
Copyright (C) 1996-98  Paul Toth (tothpaul@mygale.org)

http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)

Interface

procedure StartFM; { start playing music         }
procedure PlayFM;  { you have to call this 100/s }
procedure StopFM;  { stop playing music }

procedure InstallTimer;
procedure RemoveTimer;

Const
 VoiceCount:byte=0;

Var
 Voice:array[0..8] of PChar; { Loop at PlayFM }

Implementation

Const
 IO=$388;
 NoSound=63;
 Op:array[0..8] of byte = ($00,$01,$02,$08,$09,$0A,$10,$11,$12);

 Instruments:array['A'..'N',0..10] of byte=(
  {baselc01}(  0,  1, 35,  0,193,243,238,222,  0,  0, 10),
  {bellong }(100, 62,219,192,255,243,  1, 98,  0,  0,  4),
  {bells   }(  7, 18, 79,  0,242,242, 96,114,  0,  0,  8),
  {crunchy1}(  0,  0,192,  0,255,255,  5,  5,  0,  3,  0),
  {elecvibe}( 19,145,151,128,154,155, 18, 17,  2,  0, 14),
  {low     }(178,177,  3,  3,145,145, 42, 42,  0,  0,  9),
  {marimba }( 18, 17, 21,128,240,240,  5,  3,  0,  0, 10),
  {org10   }( 36, 49, 79,  0,  2, 82, 11, 11,  0,  0, 14),
  {organfd1}( 36, 33,  0,  0,143,247, 15,  9,  1,  1,  7),
  {pipes1  }( 32, 34, 91,128,  0, 80, 22, 21,  0,  0, 10),
  {slitdrum}( 18, 17,144,128,240,240,  7,  5,  0,  0,  0),
  {steeld  }(242,177,192,  0, 52,162,132,246,  0,  0,  4),
  {vibrells}( 82,  1, 79,  0,241,242, 83,116,  0,  0,  6),
  {voxsynth}( 65, 65, 79,  0,241,242,115,116,  1,  0,  6)
 );

 Note:array['a'..'l'] of word = (
  $16B,$181,$198,$1B0,$1CA,$1E5,$202,$220,$241,$263,$287,$2AE
{  $157,$16B,$181,$198,$1B0,$1CA,$1EA,$202,$220,$241,$263,$287 }
 );

 busy:boolean=false;

Type
 instrument=record
  Sound   :word;
  Level   :word;
  ATDC    :word;
  STRL    :word;
  Wave    :word;
  Feedback:Byte;
 end;

Const
 Active :boolean=false;
Var
 Index  :array[0..8] of word;
 Timer  :array[0..8] of integer;
 Octave :array[0..8] of byte;
 Scale  :array[0..8] of word;
 Connect:array[0..8] of byte;

Procedure WriteFM(reg,val:byte); assembler;
 asm
  mov dx,IO; mov al,reg; out dx,al;
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  inc dx; mov al,Val; out dx,al;
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
 end;

procedure SetInstrument(Voice:byte; Instr:Char);
 var
  Operator:byte;
 begin
  Operator:=Op[Voice];
  With Instrument(Instruments[Instr]) do begin
   Scale[voice]:=level;
   Connect[Voice]:=Feedback AND 1;
  { Modulator }
   WriteFM(Operator+$20,lo(Sound));
   WriteFM(Operator+$60,lo(ATDC));
   WriteFM(Operator+$80,lo(STRL));
   WriteFM(Operator+$E0,lo(Wave));
  { Carrier }
   WriteFM(Operator+$23,Hi(Sound));
   WriteFM(Operator+$63,hi(ATDC));
   WriteFM(Operator+$83,hi(STRL));
   WriteFM(Operator+$E3,hi(Wave));
  { FeedBack }
   WriteFM(Voice+$C0,Feedback);
  end;
 end;

procedure SetVolume(Voice,Level { 63-Volume } :byte);
 var
  Reg:byte;
  KSL:byte;
 begin
  Reg:=$40+Op[Voice];
  KSL:=lo(scale[voice]);
  if Connect[Voice]=1 then KSL:=(KSL and $C0) or level;
  WriteFM(Reg,KSL);
  KSL:=hi(scale[voice]) and $C0;
  WriteFM(Reg+3,KSL or level);
 end;


procedure PlayFM; { 100/s }
 var
  v:integer;
  s:word;
  p:pchar;
  c:char;
 begin
  if (busy) or (not Active) then exit;

  busy:=true;

  for v:=0 to VoiceCount-1 do begin

   if timer[v]=0 then begin

    s:=index[v];
    p:=voice[v];
    c:=p[s]; inc(s);

   { StopSound }
    WriteFM(v+$B0,0);

   { End of Partition }
    if c=#0 then begin
     inc(Timer[v]);
     continue;
    {
     c:=p[1];  -- skip first char (intrument)
      s:=2;
    }
    end;

   { Restart at \xxx }
    if c='\' then begin
     s:=ord(p[s]);
     c:=p[s]; inc(s);
    end;

   { Change Instrument }
    if c in ['A'..'Z'] then begin
     SetInstrument(v,c);
     c:=p[s]; inc(s);
    end;

   { Select Octave }
    if c<#8 then begin
     Octave[v]:=(ord(c) shl 2) or $20;
     c:=p[s]; inc(s);
    end;

   { Select Volume }
    if c<#64 then begin
     SetVolume(v,63-ord(c));
     c:=p[s]; inc(s);
    end;

   { Silence or Note }
    if c>'l' then
     SetVolume(v,NoSound)
    else begin
     WriteFM(v+$A0,(Note[c] and $FF));
     WriteFM(v+$B0,(Note[c] shr $08) or Octave[v]);
    end;

   { Delay on 1 or 2 byte }
    c:=p[s]; inc(s);
    if c<#255 then
     Timer[v]:=ord(c)
    else begin
     Timer[v]:=255+ord(p[s]); inc(s);
    end;

    index[v]:=s;

   end; { timer=0 }

   if timer[v]>0 then dec(timer[v]);

  end; { next voice }

  busy:=false;
 end;

procedure StartFm;
 begin
  WriteFM($08,$00);
  WriteFM($01,$20);
  WriteFM($BD,$DF);
  fillchar(index,sizeof(index),0);
  fillchar(timer,sizeof(timer),0);
  Active:=(VoiceCount>0);
 end;

procedure StopFm;
 var
  v:byte;
 begin
  Active:=false;
  for v:=0 to VoiceCount-1 do begin
   SetVolume(v,NOSOUND);
   WriteFM($B0+v,0);
  end;
 end;

{ Timer }
Const
 TimerInt=8;
 freq=100;
 counter=$1234DD div freq;

Var
 Ticks:word;
 OldTimer:pointer;

procedure TimerHandler; interrupt; assembler;
 asm
  add ticks,counter
  jnc @not_now
   pushf
   call oldTimer
   jmp @done
  @not_now:
   mov dx,$20
   mov al,dl
   out dx,al
  @done:
  call playfm
 end;

procedure installtimer; assembler;
 asm
  mov Ticks,0

  mov ax,3500h + TimerInt { get old Timer }
  int 21h
  mov word(OldTimer+0),bx
  mov word(OldTimer+2),es

 push ds                  { set new Timer }
  mov ax,cs
  mov ds,ax
  mov dx,offset TimerHandler
  mov ax,2500h + TimerInt
  int 21h
 pop  ds

  mov dx,$43   { setup timer }
  mov al,$34
  out dx,al

  mov dx,$40
  mov ax,Counter
  out dx,al
  mov al,ah
  out dx,al
 end;

procedure removetimer; assembler;
 asm
  mov dx,$43    { reset timer }
  mov al,$34
  out dx,al

  mov dx,$40
  xor al,al
  out dx,al
  out dx,al

  push ds        { restore Timer }
   lds dx,OldTimer
   mov ax,2500h + TimerInt
   int 21h
  pop  ds
 end;


end.