Unit Sprites;

{$i-}

Interface

Function LoadSpriteLib(LibName:string):boolean;
Procedure SaveSpriteLib(LibName:string);

Function SpriteCount:word;

Procedure DrawSprite(index,x,y:word);

Procedure ReplaceSprite(Index:word; var Data; DataSize:word);
Procedure InsertSprite(Index:word; var Data; DataSize:word);
Procedure AppendSprite(var Data; DataSize:word);
Procedure RemoveSprite(Index:word);

Procedure Encode(Width,Height:word; var Data; index:word);

Procedure Snap(sx,sy,Width,Height,index:word);

Type
 pSpriteLib=^TSpriteLib;
 TSpriteLib=record
  Count  :word;
  DataOfs:array[1..500] of word; { [1..Count] }
 {Data   :array[1..Count] of SpriteData;}
 end;

Const
 LibSize:word=0;
 SpriteHook:pointer=ptr($A000,0);

Var
 SpriteLib:pSpriteLib;

 TempSize:word;
 TempLib :pSpriteLib;


Implementation

Function LoadSpriteLib(LibName:string):boolean;
 var
  f:file;
 begin
  LoadSpriteLib:=False;

  assign(f,LibName);
  reset(f,1);
  if IoResult<>0 then exit;

  if LibSize<>0 then FreeMem(SpriteLib,LibSize);

  LibSize:=FileSize(f);
  GetMem(SpriteLib,LibSize);

  BlockRead(f,SpriteLib^,LibSize);
  LoadSpriteLib:=(IoResult=0);

  close(f);
 end;

Procedure SaveSpriteLib(LibName:string);
 var
  f:file;
 begin
  assign(f,LibName);
  rewrite(f,1);
  if ioresult<>0 then exit;
  BlockWrite(f,SpriteLib^,LibSize);
  close(f);
 end;

Function SpriteCount:word;
 begin
  if LibSize=0 then
   SpriteCount:=0
  else
   SpriteCount:=SpriteLib^.Count;
 end;

Procedure DrawSprite(index,x,y:word); assembler;
 Const
  By320:word=320;
 asm
 push ds
  les di,SpriteHook

  mov ax,y
  mul By320
  add ax,x
  add di,ax

  lds si,SpriteLib
  mov bx,Index
  add bx,bx
  add si,[si+bx]

  xor ch,ch
  lodsw
 @Continue:
  add di,ax
  lodsb
  mov cl,al
  rep movsb
  lodsw
  or  ax,ax
 jnz @Continue
 pop ds
 end;

Function SpriteSize(Index:word):word;
 begin
  SpriteSize:=0;
  if (LibSize=0)or(SpriteLib^.Count<Index) then exit;

  With SpriteLib^ do
   if Index=Count then
    SpriteSize:=LibSize-DataOfs[Index]
   else
    SpriteSize:=DataOfs[Index+1]-DataOfs[Index];
 end;

Procedure ReplaceSprite(Index:word; var Data; DataSize:word);
 var
  src,dst:^byte;
  x,s:word;
 begin
  If (LibSize=0)or(SpriteLib^.Count<Index) then exit;

  TempSize:=LibSize+DataSize-SpriteSize(Index);

  GetMem(TempLib,TempSize);

  src:=pointer(SpriteLib);
  dst:=pointer(TempLib);

 { Copy Count & Offsets }
  x:=2+2*SpriteCount;
  Move(Src^,Dst^,x);
  inc(Src,x);
  inc(Dst,x);

 { Copy Data for 1..Index-1 }
  For s:=1 to Index-1 do begin
   x:=SpriteSize(s);
   Move(Src^,Dst^,x);
   inc(Src,x);
   inc(Dst,x);
  end;

 { Replace Index }
  Move(Data,Dst^,DataSize);
  Inc(Dst,DataSize);
  Inc(Src,SpriteSize(Index));

 { Copy and Adjust for Index+1..SpriteCount }
  For s:=Index+1 to SpriteCount do begin
   x:=SpriteSize(s);
   Move(Src^,Dst^,x);
   TempLib^.DataOfs[s]:=Ofs(Dst^)-Ofs(TempLib^);
   inc(Src,x);
   inc(Dst,x);
  end;

  FreeMem(SpriteLib,LibSize);
  SpriteLib:=TempLib;
  LibSize:=TempSize;

 end;

Procedure InsertSprite(Index:word; var Data; DataSize:word);
 var
  src,dst:^byte;
  x,s:word;
 begin
  If (LibSize=0)or(SpriteLib^.Count<Index) then begin
   AppendSprite(Data,DataSize);
   exit;
  end;

  TempSize:=LibSize+DataSize+2;

  GetMem(TempLib,TempSize);

  src:=pointer(SpriteLib);
  dst:=pointer(TempLib);

 { Copy Count & Offsets }
  x:=2+SpriteLib^.Count*2;
  Move(Src^,Dst^,x);
  inc(Src,x);
  inc(Dst,x+2); { reserve place for a new offset }

 { Copy and Adjust for 1..Index-1 }
  For s:=1 to Index-1 do begin
   x:=SpriteSize(s);
   Move(Src^,Dst^,x);
   TempLib^.DataOfs[s]:=Ofs(Dst^)-Ofs(TempLib^);
   inc(Src,x);
   inc(Dst,x);
  end;

 { Insert Index }
  Inc(TempLib^.Count);
  Move(Data,Dst^,DataSize);
  TempLib^.DataOfs[Index]:=Ofs(Dst^)-Ofs(TempLib^);
  Inc(Dst,DataSize);

 { Copy and Adjust for Index(+1)..SpriteCount(+1) }
  For s:=Index to SpriteCount do begin
   x:=SpriteSize(s);
   Move(Src^,Dst^,x);
   TempLib^.DataOfs[s+1]:=Ofs(Dst^)-Ofs(TempLib^);
   inc(Src,x);
   inc(Dst,x);
  end;

  FreeMem(SpriteLib,LibSize);
  SpriteLib:=TempLib;
  LibSize:=TempSize;

 end;

Procedure AppendSprite(var Data; DataSize:word);
 var
  src,dst:^byte;
  x,s:word;
 begin
 { Define TempSize }
  If (LibSize=0) then begin
   TempSize:=2+2+DataSize;
  end else
   TempSize:=LibSize+2+DataSize;

  GetMem(TempLib,TempSize);

  if LibSize=0 then begin
  { First sprite }
   TempLib^.Count:=1;
   x:=2+2;
  end else begin

   src:=pointer(SpriteLib);
   dst:=pointer(TempLib);
  { Copy Count & Offsets }
   x:=2+2*SpriteCount;
   Move(src^,dst^,x);
   inc(src,x);
   inc(dst,x+2); { Reserve place for a new offset }

  { Copy and Adjust for 1..SpriteCount }
   For s:=1 to SpriteCount do begin
    x:=SpriteSize(s);
    Move(Src^,Dst^,x);
    TempLib^.DataOfs[s]:=Ofs(Dst^)-Ofs(TempLib^);
    inc(Src,x);
    inc(Dst,x);
   end;

   Inc(TempLib^.Count);

   FreeMem(SpriteLib,LibSize);

   x:=LibSize+2;
  end;

  { Append new sprite }
  TempLib^.DataOfs[TempLib^.Count]:=x;
  dst:=pointer(TempLib);
  inc(dst,x);
  Move(Data,dst^,DataSize);

  SpriteLib:=TempLib;
  LibSize:=TempSize;
 end;

Procedure RemoveSprite(Index:word);
 var
  src,dst:^byte;
  x,s:word;
 begin
  If (LibSize=0)or(SpriteLib^.Count<Index) then exit;

  TempSize:=LibSize-SpriteSize(Index);

  GetMem(TempLib,TempSize);

  src:=pointer(SpriteLib);
  dst:=pointer(TempLib);

  x:=2+(SpriteLib^.Count-1)*2;
  Move(Src^,Dst^,x);
  inc(Src,x+2);
  inc(Dst,x);

  For s:=1 to Index-1 do begin
   x:=SpriteSize(s);
   Move(Src^,Dst^,x);
   inc(Src,x);
   inc(Dst,x);
  end;

  For s:=Index+1 to SpriteLib^.Count do begin
   x:=SpriteSize(s);
   Move(Src^,Dst^,x);
   TempLib^.DataOfs[s]:=Ofs(Dst^)-Ofs(TempLib^);
   inc(Src,x);
   inc(Dst,x);
  end;

  FreeMem(SpriteLib,LibSize);
  SpriteLib:=TempLib;
  LibSize:=TempSize;

 end;

Procedure Encode(Width,Height:word; var Data; Index:word);
 type
  pbyte=^byte;
 var
  src:pbyte;
  dst:pbyte;
  spr:pointer;
  size  :word;
  x,y   :word;
  nuls  :word;
  pixels:byte;
  count :pbyte;
 begin
  getmem(spr,65000);
  dst:=spr;
  src:=addr(data);

  nuls:=0;
  pixels:=0;
  size:=0;

  for y:=1 to height do begin
   for x:=1 to width do begin

    if src^=0 then begin

     if (nuls=0)and(size>0) then begin
      count^:=pixels;
      pixels:=0;
     end;
     inc(nuls);

    end else begin


     if (nuls>0)or(size=0) then begin
      dst^:=nuls and $ff; inc(dst); inc(size);
      dst^:=nuls shr 8;   inc(dst); inc(size);
      nuls:=0;
      count:=dst; inc(dst); inc(size);
      pixels:=0;
     end;
     dst^:=src^; inc(dst); inc(pixels);
     inc(size);

    end;

    inc(src);

   end; { x }

   if nuls>0 then
    inc(nuls,320-width)
   else begin
    count^:=pixels; pixels:=0;
    nuls:=320-width;
   end;

  end; { y }

  if size>0 then begin

   dst^:=0; inc(dst); inc(size);
   dst^:=0; inc(dst); inc(size);

   if Index<=SpriteCount then
    ReplaceSprite(Index,spr^,size)
   else
    InsertSprite(Index,spr^,size);

  end;

  freemem(spr,65000);

 end;

Procedure Snap(sx,sy,width,height,index:word);
 type
  pbyte=^byte;
 var
  src:word;
  dst:pbyte;
  spr:pointer;
  size  :word;
  x,y   :word;
  nuls  :word;
  pixels:byte;
  count :pbyte;
 begin
  getmem(spr,65000);
  dst:=spr;

  nuls:=0;
  pixels:=0;
  size:=0;

  for y:=1 to height do begin

   src:=sx+320*sy;

   for x:=1 to width do begin

    if Mem[SegA000:src]=0 then begin

     if (nuls=0)and(size>0) then begin
      count^:=pixels;
      pixels:=0;
     end;
     inc(nuls);

    end else begin


     if (nuls>0)or(size=0) then begin
      dst^:=nuls and $ff; inc(dst); inc(size);
      dst^:=nuls shr 8;   inc(dst); inc(size);
      nuls:=0;
      count:=dst; inc(dst); inc(size);
      pixels:=0;
     end;
     dst^:=Mem[SegA000:src]; inc(dst); inc(pixels);
     inc(size);

    end;

    inc(src);

   end; { x }

   inc(sy);

   if nuls>0 then
    inc(nuls,320-width)
   else begin
    count^:=pixels; pixels:=0;
    nuls:=320-width;
   end;

  end; { y }

  if size>0 then begin

   dst^:=0; inc(dst); inc(size);
   dst^:=0; inc(dst); inc(size);

   if Index<=SpriteCount then
    ReplaceSprite(Index,spr^,size)
   else
    InsertSprite(Index,spr^,size);

  end;

  freemem(spr,65000);

 end;


end.