Unit GameKit;

{ This game kit is designed to create a game like ... Zelda ? }

(*
Copyright (C) 1996-98  Paul Toth (tothpaul@mygale.org)

http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)

{$DEFINE JOYSTICK}

Interface

Uses
 AdlibFM;

Const
 Top     = 8;
 Left    = 5;
 TopLeft = 320*Top+Left;

Type
 TPalette=array[0..255,0..2] of byte; { 256*RGB }
 TSprite =array[1..16,1..16] of byte;
 TOffset =array[1..11,1..13] of word;
 TSprites=array[1..143] of TSprite;

 TSpriteLib=record
  hook   :TOffset;
  Canvas :TOffset;
  Sprites:TSprites;
 end;

 TWorld  =array[0..11*20-1,0..13*20-1] of Byte;

 TLines  =array[1..11*16] of word;

 TScreen =array[0..319,0..199] of byte;

Var
 Palette   :TPalette;    { RGB Colors }
 SpriteLib :^TSpriteLib;
 Canevas   :^TOffset;
 Tiles     :^TSprites;   { All the sprites }
 World     :^TWorld;     { Pointer to the complet world }

 BackGround:^TScreen; { The screen without sprites }
 SwapScreen:^TScreen; { The screen with sprites }

 delay   :word;

procedure InitGameKit;
procedure InitGraphic;
Procedure BuildBackground(wx,wy:word);
Procedure RestoreBackground;
Procedure ShowScreen;
{Procedure DrawSprite(Index,x,y:integer);}
Procedure DoneGameKit;
{$ifdef joystick}
procedure ReadJoystick;
{$endif}
var
 Keyboard:array[1..128] of boolean;

Const
 key_esc  =01;
 key_left =75;
 key_right=77;
 key_up   =72;
 key_down =80;
 key_ctrl =29;
 key_alt  =56;
{$IFDEF JOYSTICK}
Procedure InitJoyStick;
Function JoyUp:Boolean;
Function JoyDown:Boolean;
Function JoyLeft:Boolean;
Function JoyRight:Boolean;
Function JoyA:boolean;
Function JoyB:boolean;
Function JoyC:boolean;
Function JoyD:boolean;
{$ENDIF}
Implementation

Var
 OldTimer:pointer;
 Old09:pointer;
 OldExit:pointer;

const
 ticks:word=0;
 freq=100;
 counter=$1234DD div freq;

 GoodBye:array[1..30] of char='Thank you for using GameKit'#13#10'$';

Const TimerInt=$08;

Procedure TimerHandler; interrupt; assembler;
 asm
  inc delay
  add ticks,counter
  jnc @not_now
   pushf
   call oldTimer
   jmp @done
  @not_now:
   mov dx,$20
   mov al,dl
   out dx,al
  @done:
  call PlayFM
 end;

Procedure KeyboardHandler; interrupt; assembler;
 asm
  mov dx,$60 { Key:=Port[$60] }
  in  al,dx
  mov bl,al

  inc dx
  in  al,dx  { dummy:=port[$61] }
  mov ah,al
  or  al,$80 { port[$61]:=dummy or $80 }
  out dx,al
  mov al,ah
  out dx,al  { port[$61]:=dummy }

  mov al,bl
  neg al     { invert press/release }
  and al,$80 { keep press/release }
  and bx,$7f
  mov [bx+offset keyboard-1],al

  mov dx,$20 { port[$20]:=$20 }
  mov al,dl
  out dx,al
 end;

{$IFDEF JOYSTICK}
Var
 JoyFound   :boolean;

 Joybuttons :byte;
 JoyAX,JoyAY:word;
 JoyBX,JoyBY:word;

 JoyMinX:word;
 JoyMaxX:word;
 JoyMinY:word;
 JoyMaxY:word;

procedure ReadJoystick; assembler; { code from PC64 }
 asm
 cli
  cmp JoyFound,False
  je  @TimeOut
  mov dx,201h
  in  al,dx          { read buttons and idle state }
 jmp @Flush
 @Flush:
  out dx,al         { start measurement }
  xor ah,ah
  mov JoyButtons,al
  and al,00001111b
  mov bl,al         { BL = idle state }
  mov cx,10

  @WaitInit:
   in AL,61h         { wait until counters are on }
  loop @WaitInit

  mov ah,00001111b  { all counters are initially on }
  @Next:
   inc cx            { increment loop counter }
   jz @TimeOut       { timeout at 65536 loops }
   in AL,DX          { read state }
   and AL,00001111b
   xor AL,AH         { AL = bits who have changed }
  jz @Next
   test AL,0001b     { check directions and store loop counter }
   je @NoAX
   mov JoyAX,CX
  @NoAX:
   test AL,0010b
   je @NoAY
   mov JoyAY,CX
  @NoAY:
   test AL,0100b
   je @NoBX
   mov JoyBX,CX
  @NoBX:
   test AL,1000b
   je @NoBY
   mov JoyBY,CX
  @NoBY:
   xor AH,AL         { set new state }
   cmp AH,BL         { idle state reached? }
  jne @Next

 jmp @Return
 @TimeOut:
   mov JoyAX,0
   mov JoyAY,0
   mov JoyBX,0
   mov JoyBY,0
  @Return:
  sti
 end;

Procedure InitJoyStick;
 Var
  Status:byte;
  dx,dy:word;
 begin
  JoyButtons:=$ff;
  JoyMinX:=0;
  JoyMaxX:=0;
  JoyMinY:=0;
  JoyMaxY:=0;

  Status:=Port[$201];
  JoyFound:=(Status=$FC)or(Status=$F0);

  if JoyFound then begin
   ReadJoystick;
   JoyFound:=(JoyAX>0)and(JoyAY>0);
   if JoyFound then begin
    dx:=(5*JoyAX) div 16;
    dy:=(5*JoyAY) div 16;
    JoyMinX:=JoyAX-dx;
    JoyMaxX:=JoyAX+dx;
    JoyMinY:=JoyAY-dy;
    JoyMaxY:=JoyAY+dy;
   end;
  end;

 end;

Function JoyUp:Boolean;    begin JoyUp:=JoyAY<JoyMinY; end;
Function JoyDown:Boolean;  begin JoyDown:=JoyAY>JoyMaxY; end;
Function JoyLeft:Boolean;  begin JoyLeft:=JoyAX<JoyMinX; end;
Function JoyRight:Boolean; begin JoyRight:=JoyAX>JoyMaxX; end;
Function JoyA:boolean; begin JoyA:=(JoyButtons and $10)=0 end;
Function JoyB:boolean; begin JoyB:=(JoyButtons and $20)=0 end;
Function JoyC:boolean; begin JoyC:=(JoyButtons and $40)=0 end;
Function JoyD:boolean; begin JoyD:=(JoyButtons and $80)=0 end;

{$ENDIF}
procedure InitGameKit; assembler;
 asm
  les di,SpriteLib      { Setup Hook }
  mov ax,TopLeft        { Top Left of the Map }
  add ax,word(Background)
  mov dl,11
 @NextRow:
  mov cx,13
 @NextCol:
  stosw
  add ax,16
 loop @NextCol
  add ax,320*16-13*16
  dec dl
 jnz @NextRow

  push ds
  pop  es

  mov di,offset Keyboard { Setup Keyboard }
  mov al,False
  mov cx,128
  rep stosb

{ Exit Proc }
  les di,ExitProc
  mov word(OldExit),di
  mov word(OldExit+2),es
  mov si,offset DoneGameKit
  mov Word(ExitProc),si
  mov Word(ExitProc+2),cs

  mov ax,3500h + TimerInt { get IV Timer }
  int 21h
  mov word(OldTimer),bx
  mov word(OldTimer+2),es

 push ds                  { set IV Timer }
  mov ds,Word(ExitProc+2)
  mov dx,offset TimerHandler
  mov ax,2500h + TimerInt
  int 21h
 pop  ds

  mov dx,$43   { setup timer }
  mov al,$34
  out dx,al

  mov dx,$40
  mov ax,Counter
  out dx,al
  mov al,ah
  out dx,al

{ keyboard }
  mov ax,3509h { get IV Keyboard }
  int 21h
  mov word(Old09),bx
  mov word(Old09+2),es

 push ds
  mov ds,Word(ExitProc+2)
  mov dx,offset keyboardhandler
  mov ax,2509h { set IV Keyboard }
  int 21h
 pop  ds
 end;

procedure InitGraphic; assembler;
 asm
  mov ax,13h; int 10h     { Mode 320x200x256 colors }
  mov dx,$3c8             { Set Palette }
  mov cx,3*256
  mov si,Offset Palette
  mov al,0
  out dx,al
  inc dx
 @NextDAC:
  lodsb
  out dx,al
 loop @NextDAC
 end;

Procedure BuildBackground(wx,wy:word); assembler;
 asm
{ Build Canvas }
 push ds

  les di,Canevas

  mov ax,wy
  mov bx,13*20
  mul bx
  add ax,wx

  mov bx,word(Tiles)

  lds si,World
  add si,ax

  mov cl,11
 @Next2:
  mov ch,13
 @Next1:
  xor ah,ah
  lodsb
  dec al
  shl ax,8 { = * SizeOf(TSprite) = * (16*16) }
  add ax,bx
  stosw
  dec ch
 jnz @Next1
  add si,13*20-13
 loop @Next2

 pop ds

{ Use Canvas ! }
  les di,Background
  mov cx,320*200/2
  xor ax,ax
  rep stosw


 push ds
  lds bx,SpriteLib

  mov dl,11*13
 @NextSprite:

  mov si,[bx+11*13*2] { canevas }
  mov di,[bx]         { hook }
  inc bx
  inc bx

  mov dh,16
 @NextLine:
  mov cx,16/2
  rep movsw
  add di,320-16
  dec dh
 jnz @NextLine

  dec dl
 jnz @NextSprite

 pop ds
 end;

Procedure RestoreBackGround; assembler;
 asm
  push ds
   les di,SwapScreen
   lds si,BackGround
   mov cx,320*200/2
   rep movsw
  pop  ds
 end;

Procedure ShowScreen; assembler;
 asm
 push ds
  mov es,SegA000
  xor di,di
  lds si,SwapScreen
  mov cx,320*200/2
  rep movsw
 pop  ds
 end;

Procedure DoneGameKit; assembler;
 asm
  call StopFM

  les di,OldExit { Restore ExitProc }
  mov word(ExitProc),di
  mov word(ExitProc+2),es

  mov dx,$43    { reset timer }
  mov al,$34
  out dx,al

  mov dx,$40
  xor al,al
  out dx,al
  out dx,al

  push ds        { restore Timer }
   lds dx,OldTimer
   mov ax,2500h + TimerInt
   int 21h
  pop  ds

  push ds        { restore Keyboard }
   lds dx,Old09
   mov ax,2509h
   int 21h
  pop  ds

  mov ax,03h    { restore Video }
  int 10h

  mov ah,09h
  mov dx,offset GoodBye
  int 21h
 end;

Const
 Dwarfs:Pchar=(
  'I'#3#63+'a'#25+'f'#30+
  'G'#3#63+'f'#25+'f'#25+
  'I'#3#63+'e'#25+'c'#25+
  'C'#3#53+'c'#25+'c'#25+
  'I'#3#63+
  'f'#25+'h'#30+'j'#15+'h'#20+'f'#25+'e'#35+
  'z'#45#60+
  'a'#25+'c'#25+'f'#25+'a'#25+
  'a'#25+'c'#25+'f'#25+'a'#25+
  'k'#25+'j'#25+
  'z'#50#63+
  'f'#25+'h'#25+
  'a'#25+'c'#25+'e'#25+
  '\'#5
 );

begin
 New(World);
 New(SpriteLib);
  Canevas:=Ptr(Seg(SpriteLib^),Ofs(SpriteLib^)+1*SizeOf(TOffset));
  Tiles  :=Ptr(Seg(SpriteLib^),Ofs(SpriteLib^)+2*SizeOf(TOffset));
 New(Background);
 New(SwapScreen);

 VoiceCount:=1;
 Voice[0]:=Dwarfs;
 StartFM;
end.